import { config } from "./config";
import { _name, _type, _validate } from "./symbols";
import { enyToGuard, enyToGuardFn } from "./utils";
import { valit } from "./valit";
import { vality } from "./vality";
vality.array = valit("array", (e) => (value, options, path) => {
    const fn = enyToGuardFn(e);
    if (!Array.isArray(value)) {
        if (!config.strict) {
            const res = fn(value, [...path, 0], value);
            if (res.valid) {
                return { valid: true, data: [res.data], errors: [] };
            }
        }
        return {
            valid: false,
            data: undefined,
            errors: [{ message: "vality.array.base", path, options, value }],
        };
    }
    const data = [];
    const errors = [];
    for (let k = 0; k < value.length; k++) {
        const res = fn(value[k], [...path, k], value);
        if (!res.valid) {
            errors.push(...res.errors);
            if (options.bail)
                break;
        }
        else {
            data.push(res.data);
        }
    }
    if (errors.length === 0)
        return { valid: true, data, errors: [] };
    return { valid: false, data: undefined, errors };
}, {
    minLength: (val, o) => val.length >= o,
    maxLength: (val, o) => val.length <= o,
}, {
    bail: false,
});
vality.object = valit("object", (e) => (value, options, path) => {
    if (typeof value !== "object" || value === null)
        return {
            valid: false,
            data: undefined,
            errors: [{ message: "vality.object.base", path, options, value }],
        };
    const data = {};
    const errors = [];
    for (const k in e) {
        const ek = e[k];
        if (typeof ek === "function" &&
            ek?.[_type] === "readonly")
            continue;
        const res = enyToGuardFn(ek)(value[k], [...path, k], value);
        if (!res.valid) {
            errors.push(...res.errors);
            if (options.bail)
                break;
        }
        else {
            data[k] = res.data;
        }
    }
    for (const k in value) {
        const ek = e[k];
        if (ek === undefined ||
            (typeof ek === "function" &&
                ek?.[_type] === "readonly")) {
            errors.push({
                message: "vality.object.extraProperty",
                path: [...path, k],
                options,
                value: value[k],
            });
            if (options.bail)
                break;
        }
    }
    if (errors.length === 0)
        return { valid: true, data: data, errors: [] };
    return { valid: false, data: undefined, errors };
}, {}, {
    bail: false,
});
vality.optional = valit("optional", (e) => (val, _options, path, parent) => {
    const enyVal = enyToGuardFn(e)(val, path, parent);
    if (enyVal.valid)
        return enyVal;
    if (val === undefined)
        return { valid: true, data: undefined, errors: [] };
    if (!config.strict && val === null)
        return { valid: true, data: undefined, errors: [] };
    return enyVal;
});
vality.enum = valit("enum", (...es) => (value, options, path, parent) => {
    for (const e of es) {
        const res = enyToGuardFn(e)(value, path, parent);
        if (res.valid)
            return res;
    }
    return {
        valid: false,
        data: undefined,
        errors: [{ message: "vality.enum.base", path, options, value }],
    };
});
vality.tuple = valit("tuple", (...es) => (value, options, path) => {
    if (!Array.isArray(value))
        return {
            valid: false,
            data: undefined,
            errors: [{ message: "vality.tuple.base", path, options, value }],
        };
    const data = [];
    const errors = [];
    for (let i = 0; i < es.length; i++) {
        const res = enyToGuardFn(es[i])(value[i], [...path, i], value);
        if (!res.valid) {
            errors.push(...res.errors);
            if (options.bail)
                break;
        }
        else {
            data[i] = res.data;
        }
    }
    for (let i = es.length; i < value.length; i++) {
        errors.push({
            message: "vality.tuple.extraProperty",
            path: [...path, i],
            options,
            value: value[i],
        });
        if (options.bail)
            break;
    }
    if (errors.length === 0)
        return { valid: true, data, errors: [] };
    return { valid: false, data: undefined, errors };
}, {}, {
    bail: false,
});
vality.readonly = valit("readonly", (e) => (val, _options, path) => {
    if (val === undefined)
        return { valid: true, data: undefined, errors: [] };
    return {
        valid: false,
        data: undefined,
        errors: [
            {
                message: "vality.readonly.base",
                path,
                options: {},
                value: val,
            },
        ],
    };
});
vality.and = valit("and", (...es) => (value, options, path, parent) => {
    if (typeof value !== "object" || value === null)
        return {
            valid: false,
            data: undefined,
            errors: [{ message: "vality.and.base", path, options, value }],
        };
    const data = {};
    const errors = [];
    const handleEs = (ess) => {
        for (let i = 0; i < ess.length; i++) {
            const eGuard = enyToGuard(ess[i]);
            const typeOfGuard = eGuard[_name];
            let res = undefined;
            switch (typeOfGuard) {
                case "object": {
                    const objectValue = {};
                    for (const k in eGuard[_validate][_type][0]) {
                        Object.assign(objectValue, {
                            [k]: value[k],
                        });
                    }
                    res = eGuard[_validate](objectValue, path, parent);
                    break;
                }
                case "enum":
                    for (const e of eGuard[_validate][_type]) {
                        const enumMemberGuardFn = enyToGuardFn(e);
                        const enumMemberValue = {};
                        for (const k in enumMemberGuardFn[_type][0]) {
                            Object.assign(enumMemberValue, {
                                [k]: value[k],
                            });
                        }
                        res = enumMemberGuardFn(enumMemberValue, path, parent);
                        if (res.valid)
                            break;
                    }
                    if (!res)
                        res = {
                            valid: false,
                            data: undefined,
                            errors: [
                                { message: "vality.enum.base", path, options, value },
                            ],
                        };
                    break;
                case "and":
                    handleEs(eGuard[_validate][_type]);
                    break;
                default:
                    throw new Error("vality.and: Unexpected type of guard: " + typeOfGuard);
            }
            if (!res)
                continue;
            if (!res.valid) {
                errors.push(...res.errors);
                if (options.bail)
                    break;
            }
            else {
                Object.assign(data, res.data);
            }
        }
    };
    handleEs(es);
    if (errors.length !== 0)
        return { valid: false, data: undefined, errors };
    const gotKeys = Object.keys(data).length;
    const expectedKeys = Object.keys(value).length;
    if (gotKeys < expectedKeys) {
        return {
            valid: false,
            data: undefined,
            errors: [
                { message: "vality.and.extraProperties", path, options, value },
            ],
        };
    }
    else if (gotKeys > expectedKeys) {
        throw new Error("This can't happen");
    }
    return { valid: true, data, errors: [] };
}, {}, {
    bail: false,
});
vality.dict = valit("dict", (k, v) => (value, options, path, parent) => {
    if (typeof value !== "object" || value === null)
        return {
            valid: false,
            data: undefined,
            errors: [{ message: "vality.dict.base", path, options, value }],
        };
    const keyGuard = enyToGuard(k);
    const typeOfKey = keyGuard[_name];
    if (typeOfKey === "literal" ||
        typeOfKey === "string" ||
        typeOfKey === "number") {
        return vality
            .dict(vality.enum(keyGuard), v)(options)[_validate](value, path, parent);
    }
    const keysGuards = keyGuard[_validate][_type].map(enyToGuard);
    const literalKeys = keysGuards.filter((g) => g[_name] === "literal");
    const errors = [];
    for (const key in value) {
        if (!keyGuard[_validate](key, [...path, key], value).valid) {
            errors.push({
                message: "vality.dict.invalidProperty",
                path: [...path, key],
                options,
                value: key,
            });
            if (options.bail)
                break;
        }
    }
    if (errors.length)
        return { valid: false, data: undefined, errors };
    for (const literalKeyGuard of literalKeys) {
        if (!Object.keys(value).some((k) => literalKeyGuard[_validate](k, [...path, k], value).valid)) {
            if (!literalKeyGuard[_validate](k, path, value).valid) {
                errors.push({
                    message: "vality.dict.missingProperty",
                    path: [
                        ...path,
                        literalKeyGuard[_validate][_type][0][_type],
                    ],
                    options,
                    value: undefined,
                });
                if (options.bail)
                    break;
            }
        }
    }
    if (errors.length)
        return { valid: false, data: undefined, errors };
    const valueGuardFn = enyToGuardFn(v);
    const data = {};
    for (const key in value) {
        const res = valueGuardFn(value[key], [...path, key], value);
        if (!res.valid) {
            errors.push(...res.errors);
            if (options.bail)
                break;
        }
        else {
            data[key] = res.data;
        }
    }
    if (errors.length)
        return { valid: false, data: undefined, errors };
    return { valid: true, data, errors: [] };
}, {}, {
    bail: false,
});
//# sourceMappingURL=valits.js.map